"""
Project Trellis Devices-of-Interest list
"""

families = {
    "ECP5": {
        "devices": {
            "LFE5U-25F": {
                "packages": ["csfBGA285", "caBGA256", "caBGA381", "caBGA554", "caBGA756"]
            },
            "LFE5U-45F": {
                "packages": ["csfBGA285", "caBGA256", "caBGA381", "caBGA554", "caBGA756"]
            },
            "LFE5U-85F": {
                "packages": ["csfBGA285", "caBGA381", "caBGA554", "caBGA756"]
            },
        }
    }
}